use crate::constants::Pattern;
use crate::convert;
use crate::token::Token;
use chrono::{DateTime, Datelike, Duration, FixedOffset};
use std::cmp;
use std::cmp::PartialEq;
use std::collections::HashMap;

const FUZZY_PATTERNS: [(&Pattern, fn(FuzzyDate, &CallValues, &Rules) -> Result<FuzzyDate, ()>); 51] = [
    // KEYWORDS
    (&Pattern::Now, |c, _, _| Ok(c)),
    (&Pattern::Today, |c, _, _| c.time_reset()),
    (&Pattern::Midnight, |c, _, _| c.time_reset()),
    (&Pattern::Yesterday, |c, _, r| c.offset_unit_keyword(TimeUnit::Days, -1, r)?.time_reset()),
    (&Pattern::Tomorrow, |c, _, r| c.offset_unit_keyword(TimeUnit::Days, 1, r)?.time_reset()),
    // WEEKDAY OFFSETS
    (&Pattern::ThisWday, |c, v, _| c.offset_weekday(v.get_int(0), convert::Change::None)?.time_reset()),
    (&Pattern::PrevWday, |c, v, _| c.offset_weekday(v.get_int(0), convert::Change::Prev)?.time_reset()),
    (&Pattern::LastWday, |c, v, _| c.offset_weekday(v.get_int(0), convert::Change::Prev)?.time_reset()),
    (&Pattern::NextWday, |c, v, _| c.offset_weekday(v.get_int(0), convert::Change::Next)?.time_reset()),
    // MONTH OFFSETS
    (&Pattern::ThisMonth, |c, v, _| c.offset_month(v.get_int(0), convert::Change::None)?.time_reset()),
    (&Pattern::PrevMonth, |c, v, _| c.offset_month(v.get_int(0), convert::Change::Prev)?.time_reset()),
    (&Pattern::LastMonth, |c, v, _| c.offset_month(v.get_int(0), convert::Change::Prev)?.time_reset()),
    (&Pattern::NextMonth, |c, v, _| c.offset_month(v.get_int(0), convert::Change::Next)?.time_reset()),
    // KEYWORD OFFSETS
    (&Pattern::ThisLongUnit, |c, v, r| c.offset_unit_keyword(v.get_unit(0), 0, r)),
    (&Pattern::PrevLongUnit, |c, v, r| c.offset_unit_keyword(v.get_unit(0), -1, r)),
    (&Pattern::LastLongUnit, |c, v, r| c.offset_unit_keyword(v.get_unit(0), -1, r)),
    (&Pattern::NextLongUnit, |c, v, r| c.offset_unit_keyword(v.get_unit(0), 1, r)),
    // NUMERIC OFFSET, MINUS
    (&Pattern::MinusUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), 0 - v.get_int(0), r)),
    (&Pattern::MinusShortUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), 0 - v.get_int(0), r)),
    (&Pattern::MinusLongUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), 0 - v.get_int(0), r)),
    // NUMERIC OFFSET, PLUS
    (&Pattern::PlusUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), v.get_int(0), r)),
    (&Pattern::PlusShortUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), v.get_int(0), r)),
    (&Pattern::PlusLongUnit, |c, v, r| c.offset_unit_exact(v.get_unit(1), v.get_int(0), r)),
    // NUMERIC OFFSET, PLUS
    (&Pattern::UnitAgo, |c, v, r| c.offset_unit_exact(v.get_unit(1), 0 - v.get_int(0), r)),
    (&Pattern::LongUnitAgo, |c, v, r| c.offset_unit_exact(v.get_unit(1), 0 - v.get_int(0), r)),
    // FIRST/LAST OFFSETS
    (&Pattern::FirstLongUnitOfMonth, |c, v, _| {
        c.offset_range_month(v.get_unit(0), v.get_int(1), convert::Change::First)?
            .time_reset()
    }),
    (&Pattern::LastLongUnitOfMonth, |c, v, _| {
        c.offset_range_month(v.get_unit(0), v.get_int(1), convert::Change::Last)?
            .time_reset()
    }),
    (&Pattern::FirstLongUnitOfThisLongUnit, |c, v, _| {
        c.offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::First)?
            .time_reset()
    }),
    (&Pattern::LastLongUnitOfThisLongUnit, |c, v, _| {
        c.offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::Last)?
            .time_reset()
    }),
    (&Pattern::FirstLongUnitOfPrevLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), -1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::First)?
            .time_reset()
    }),
    (&Pattern::LastLongUnitOfPrevLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), -1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::Last)?
            .time_reset()
    }),
    (&Pattern::FirstLongUnitOfLastLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), -1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::First)?
            .time_reset()
    }),
    (&Pattern::LastLongUnitOfLastLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), -1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::Last)?
            .time_reset()
    }),
    (&Pattern::FirstLongUnitOfNextLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), 1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::First)?
            .time_reset()
    }),
    (&Pattern::LastLongUnitOfNextLongUnit, |c, v, r| {
        c.offset_unit_keyword(v.get_unit(1), 1, r)?
            .offset_range_unit(v.get_unit(0), v.get_unit(1), convert::Change::Last)?
            .time_reset()
    }),
    // @1705072948, @1705072948.452
    (&Pattern::Timestamp, |c, v, _| c.date_stamp(v.get_int(0), 0)),
    (&Pattern::TimestampFloat, |c, v, _| c.date_stamp(v.get_int(0), v.get_ms(1))),
    // 2023-01-01, 30.1.2023, 1/30/2023
    (&Pattern::DateYmd, |c, v, _| c.date_ymd(v.get_int(0), v.get_int(1), v.get_int(2))?.time_reset()),
    (&Pattern::DateDmy, |c, v, _| c.date_ymd(v.get_int(2), v.get_int(1), v.get_int(0))?.time_reset()),
    (&Pattern::DateMdy, |c, v, _| c.date_ymd(v.get_int(2), v.get_int(0), v.get_int(1))?.time_reset()),
    // Dec 7, Dec 7th, 7 Dec
    (&Pattern::DateMonthDay, |c, v, _| c.date_ymd(c.year(), v.get_int(0), v.get_int(1))?.time_reset()),
    (&Pattern::DateMonthNth, |c, v, _| c.date_ymd(c.year(), v.get_int(0), v.get_int(1))?.time_reset()),
    (&Pattern::DateDayMonth, |c, v, _| c.date_ymd(c.year(), v.get_int(1), v.get_int(0))?.time_reset()),
    // Dec 7 2023, Dec 7th 2023, 7 Dec 2023
    (&Pattern::DateMonthDayYear, |c, v, _| c.date_ymd(v.get_int(2), v.get_int(0), v.get_int(1))?.time_reset()),
    (&Pattern::DateMonthNthYear, |c, v, _| c.date_ymd(v.get_int(2), v.get_int(0), v.get_int(1))?.time_reset()),
    (&Pattern::DateDayMonthYear, |c, v, _| c.date_ymd(v.get_int(2), v.get_int(1), v.get_int(0))?.time_reset()),
    // 2023-12-07 15:02, 2023-12-07 15:02:01, 2023-12-07 15:02:01.456
    (&Pattern::DateTimeYmdHm, |c, v, _| {
        c.date_ymd(v.get_int(0), v.get_int(1), v.get_int(2))?
            .time_hms(v.get_int(3), v.get_int(4), 0, 0)
    }),
    (&Pattern::DateTimeYmdHms, |c, v, _| {
        c.date_ymd(v.get_int(0), v.get_int(1), v.get_int(2))?
            .time_hms(v.get_int(3), v.get_int(4), v.get_int(5), 0)
    }),
    (&Pattern::DateTimeYmdHmsMs, |c, v, _| {
        c.date_ymd(v.get_int(0), v.get_int(1), v.get_int(2))?.time_hms(
            v.get_int(3),
            v.get_int(4),
            v.get_int(5),
            v.get_ms(6),
        )
    }),
    // 3pm, 3:00 pm
    (&Pattern::TimeMeridiemH, |c, v, _| c.time_12h(v.get_int(0), 0, 0, v.get_int(1))),
    (&Pattern::TimeMeridiemHm, |c, v, _| c.time_12h(v.get_int(0), v.get_int(1), 0, v.get_int(2))),
];

#[derive(PartialEq)]
enum TimeUnit {
    Days,
    Hours,
    Minutes,
    Months,
    Seconds,
    Weeks,
    Years,
    None,
}

impl TimeUnit {
    fn from_int(value: i64) -> TimeUnit {
        match value {
            1 => Self::Seconds,
            2 => Self::Minutes,
            3 => Self::Hours,
            4 => Self::Days,
            5 => Self::Weeks,
            6 => Self::Months,
            7 => Self::Years,
            _ => Self::None,
        }
    }
}

struct CallValues {
    tokens: Vec<Token>,
}

impl CallValues {
    fn from_tokens(tokens: Vec<Token>) -> Self {
        Self { tokens: tokens }
    }

    fn drop_used(&mut self, used: usize) {
        self.tokens = self.tokens[used..].to_owned();
    }

    fn get_int(&self, index: usize) -> i64 {
        self.tokens[index].value
    }

    /// Get value with the assumption that it should represent milliseconds,
    /// and thus the number of zeros before the number is meaningful. If there
    /// are too many zeros, we use -1 to break out on millisecond value
    /// validation.
    fn get_ms(&self, index: usize) -> i64 {
        let value = self.tokens[index].value;
        let zeros = self.tokens[index].zeros;

        let multiply_by = if value.lt(&10) {
            match zeros {
                0 => 100,
                1 => 10,
                2 => 1,
                _ => return -1,
            }
        } else if value.lt(&100) {
            match zeros {
                0 => 10,
                1 => 1,
                _ => return -1,
            }
        } else if value.lt(&1000) {
            match zeros {
                0 => 1,
                _ => return -1,
            }
        } else {
            return -1;
        };

        value * multiply_by
    }

    fn get_unit(&self, index: usize) -> TimeUnit {
        TimeUnit::from_int(self.get_int(index))
    }
}

struct FuzzyDate {
    time: DateTime<FixedOffset>,
}

impl FuzzyDate {
    fn new(time: DateTime<FixedOffset>) -> Self {
        FuzzyDate { time: time }
    }

    /// Set time to specific timestamp
    fn date_stamp(&self, sec: i64, ms: i64) -> Result<Self, ()> {
        Ok(Self { time: convert::date_stamp(sec, ms) })
    }

    /// Set time to specific year, month and day
    fn date_ymd(&self, year: i64, month: i64, day: i64) -> Result<Self, ()> {
        Ok(Self { time: convert::date_ymd(self.time, year, month, day)? })
    }

    /// Move time into previous or upcoming month
    fn offset_month(&self, new_month: i64, change: convert::Change) -> Result<Self, ()> {
        Ok(Self { time: convert::offset_month(self.time, new_month, change) })
    }

    /// Move time into previous or upcoming weekday
    fn offset_weekday(&self, new_weekday: i64, change: convert::Change) -> Result<Self, ()> {
        Ok(Self { time: convert::offset_weekday(self.time, new_weekday, change) })
    }

    /// Move time within month range
    fn offset_range_month(&self, target: TimeUnit, month: i64, change: convert::Change) -> Result<Self, ()> {
        if target.eq(&TimeUnit::Days) {
            let new_time = convert::offset_range_month(self.time, month, change)?;
            return Ok(Self { time: new_time });
        }

        Err(())
    }

    /// Move time within unit range
    fn offset_range_unit(&self, target: TimeUnit, unit: TimeUnit, change: convert::Change) -> Result<Self, ()> {
        if !(target.eq(&TimeUnit::Days) && unit.eq(&TimeUnit::Months)) {
            return Err(());
        }

        let new_day: u32 = match change.eq(&convert::Change::Last) {
            true => convert::into_month_day(self.time.year(), self.time.month(), 32),
            false => 1,
        };

        Ok(Self { time: self.time.with_day(new_day).unwrap() })
    }

    /// Move time exactly by specified number of units
    fn offset_unit_exact(&self, target: TimeUnit, amount: i64, _rules: &Rules) -> Result<FuzzyDate, ()> {
        let new_time = match target {
            TimeUnit::Seconds => self.time + Duration::seconds(amount),
            TimeUnit::Minutes => self.time + Duration::minutes(amount),
            TimeUnit::Hours => self.time + Duration::hours(amount),
            TimeUnit::Days => self.time + Duration::days(amount),
            TimeUnit::Weeks => self.time + Duration::days(amount * 7),
            TimeUnit::Months => convert::offset_months(self.time, amount),
            TimeUnit::Years => convert::offset_years(self.time, amount),
            _ => self.time,
        };

        Ok(Self { time: new_time })
    }

    /// Move time by specific unit, but apply keyword rules where
    /// e.g. moving by weeks will land on to first day of week
    fn offset_unit_keyword(&self, target: TimeUnit, amount: i64, rules: &Rules) -> Result<FuzzyDate, ()> {
        let new_time = match target {
            TimeUnit::Weeks => convert::offset_weeks(self.time, amount, rules.week_start_day()),
            _ => return self.offset_unit_exact(target, amount, rules),
        };

        Ok(Self { time: new_time })
    }

    /// Set time to specific hour, minute and second using 12-hour clock
    fn time_12h(&self, hour: i64, min: i64, sec: i64, meridiem: i64) -> Result<Self, ()> {
        Ok(Self { time: convert::time_12h(self.time, hour, min, sec, meridiem)? })
    }

    /// Set time to specific hour, minute and second
    fn time_hms(&self, hour: i64, min: i64, sec: i64, ms: i64) -> Result<Self, ()> {
        Ok(Self { time: convert::time_hms(self.time, hour, min, sec, ms)? })
    }

    /// Reset time to midnight
    fn time_reset(&self) -> Result<Self, ()> {
        self.time_hms(0, 0, 0, 0)
    }

    /// Current year
    fn year(&self) -> i64 {
        self.time.year() as i64
    }
}

struct Rules {
    week_start_mon: bool,
}

impl Rules {
    fn week_start_day(&self) -> i8 {
        match self.week_start_mon {
            true => 1,
            false => 7,
        }
    }
}

pub(crate) struct UnitLocale {
    day: String,
    days: String,
    hour: String,
    hours: String,
    minute: String,
    minutes: String,
    second: String,
    seconds: String,
    week: String,
    weeks: String,
    separator: String,
}

impl UnitLocale {
    pub(crate) fn from_map(names: HashMap<String, String>) -> Self {
        let mut mapping: HashMap<String, String> = HashMap::from([
            (String::from("second"), String::from("")),
            (String::from("seconds"), String::from("")),
            (String::from("minute"), String::from("")),
            (String::from("minutes"), String::from("")),
            (String::from("hour"), String::from("")),
            (String::from("hours"), String::from("")),
            (String::from("day"), String::from("")),
            (String::from("days"), String::from("")),
            (String::from("week"), String::from("")),
            (String::from("weeks"), String::from("")),
        ]);

        mapping.extend(names);

        Self {
            day: mapping.get("day").unwrap().to_owned(),
            days: mapping.get("days").unwrap().to_owned(),
            hour: mapping.get("hour").unwrap().to_owned(),
            hours: mapping.get("hours").unwrap().to_owned(),
            minute: mapping.get("minute").unwrap().to_owned(),
            minutes: mapping.get("minutes").unwrap().to_owned(),
            second: mapping.get("second").unwrap().to_owned(),
            seconds: mapping.get("seconds").unwrap().to_owned(),
            week: mapping.get("week").unwrap().to_owned(),
            weeks: mapping.get("weeks").unwrap().to_owned(),
            separator: if mapping.get("day").unwrap().len() > 1 { " " } else { "" }.to_owned(),
        }
    }

    fn format_days(&self, amount: i32) -> String {
        let unit = if amount.eq(&1) { &self.day } else { &self.days };
        format!(" {}{}{}", amount, self.separator, unit)
    }

    fn format_hours(&self, amount: i32) -> String {
        let unit = if amount.eq(&1) { &self.hour } else { &self.hours };
        format!(" {}{}{}", amount, self.separator, unit)
    }

    fn format_minutes(&self, amount: i32) -> String {
        let unit = if amount.eq(&1) { &self.minute } else { &self.minutes };
        format!(" {}{}{}", amount, self.separator, unit)
    }

    fn format_seconds(&self, amount: i32) -> String {
        let unit = if amount.eq(&1) { &self.second } else { &self.seconds };
        format!(" {}{}{}", amount, self.separator, unit)
    }

    fn format_weeks(&self, amount: i32) -> String {
        let unit = if amount.eq(&1) { &self.week } else { &self.weeks };
        format!(" {}{}{}", amount, self.separator, unit)
    }
}

/// Perform conversion against pattern and corresponding token values,
/// relative to given datetime
pub(crate) fn convert(
    pattern: &str,
    tokens: Vec<Token>,
    current_time: &DateTime<FixedOffset>,
    week_start_mon: bool,
    custom_patterns: HashMap<String, String>,
) -> Option<DateTime<FixedOffset>> {
    let call_list = find_pattern_calls(&pattern, custom_patterns);

    if call_list.len().eq(&0) {
        return None;
    }

    let rules = Rules { week_start_mon: week_start_mon };

    let mut ctx_time = FuzzyDate::new(current_time.to_owned());
    let mut ctx_vals = CallValues::from_tokens(tokens);

    for (pattern_match, pattern_call) in call_list {
        ctx_time = match pattern_call(ctx_time, &ctx_vals, &rules) {
            Ok(value) => value,
            Err(_) => return None,
        };
        ctx_vals.drop_used(pattern_match.split("[").count() - 1);
    }

    Option::from(ctx_time.time)
}

/// Turn seconds into a duration string
pub(crate) fn to_duration(seconds: f64, units: &UnitLocale, max_unit: &str, min_unit: &str) -> String {
    let mut seconds = seconds;
    let mut result: String = String::from("");

    let naming: HashMap<&str, i8> = HashMap::from([
        ("s", 1),
        ("sec", 1),
        ("min", 2),
        ("mins", 2),
        ("h", 3),
        ("hr", 3),
        ("hrs", 3),
        ("d", 4),
        ("day", 4),
        ("days", 4),
        ("w", 5),
        ("week", 5),
        ("weeks", 5),
    ]);

    let max_u: &i8 = naming.get(max_unit).unwrap_or(&5);
    let min_u: &i8 = naming.get(min_unit).unwrap_or(&1);

    if max_u.ge(&5) && min_u.le(&5) {
        let weeks = (seconds / 604800.0).floor() as i32;

        if weeks.gt(&0) {
            result.push_str(&units.format_weeks(weeks));
            seconds -= (weeks * 604800) as f64;
        }
    }

    if max_u.ge(&4) && min_u.le(&4) {
        let days = (seconds / 86400.0).floor() as i32;

        if days.gt(&0) {
            result.push_str(&units.format_days(days));
            seconds -= (days * 86400) as f64;
        }
    }

    if max_u.ge(&3) && min_u.le(&3) {
        let hours = (seconds / 3600.0).floor() as i32;

        if hours.gt(&0) {
            result.push_str(&units.format_hours(hours));
            seconds -= (hours * 3600) as f64;
        }
    }

    if max_u.ge(&2) && min_u.le(&2) {
        let minutes = (seconds / 60.0).floor() as i32;

        if minutes.gt(&0) {
            result.push_str(&units.format_minutes(minutes));
            seconds -= (minutes * 60) as f64;
        }
    }

    if max_u.ge(&1) && min_u.le(&1) {
        if seconds.gt(&0.0) {
            result.push_str(&units.format_seconds(seconds as i32));
        }
    }

    result.trim().to_string()
}

/// Find closure calls that match the pattern exactly, or partially
fn find_pattern_calls(
    pattern: &str,
    custom: HashMap<String, String>,
) -> Vec<(String, fn(FuzzyDate, &CallValues, &Rules) -> Result<FuzzyDate, ()>)> {
    let mut closure_map = HashMap::new();
    let mut pattern_map = HashMap::new();

    for (pattern_type, closure_function) in FUZZY_PATTERNS {
        closure_map.insert(pattern_type, closure_function);
        for pattern_value in Pattern::values(pattern_type) {
            pattern_map.insert(pattern_value.to_string(), pattern_type.to_owned());
        }
    }

    for (custom_pattern, closure_pattern) in custom {
        if pattern_map.contains_key(&closure_pattern) {
            pattern_map.insert(custom_pattern.to_owned(), pattern_map.get(&closure_pattern).unwrap());
        }
    }

    for prefix in vec!["", "+"] {
        let try_pattern = format!("{}{}", prefix, pattern);

        if pattern_map.contains_key(&try_pattern) {
            let pattern_type = pattern_map.get(&try_pattern).unwrap();
            return vec![(try_pattern.to_string(), *closure_map.get(pattern_type).unwrap())];
        }
    }

    let mut result = vec![];
    let mut search = pattern;
    let prefix = if pattern.starts_with("-") { "-" } else { "+" };

    while search.len().gt(&0) {
        let mut calls: Vec<(&str, &Pattern)> = vec![];

        for (map_pattern, map_type) in &pattern_map {
            if search.starts_with(&map_pattern.as_str())
                || format!("{}{}", prefix, search).starts_with(&map_pattern.as_str())
            {
                calls.push((&map_pattern, map_type));
            }
        }

        if calls.len().eq(&0) {
            return vec![];
        }

        calls.sort_by(|a, b| b.0.cmp(a.0));
        let (best_match, best_type) = calls.first().unwrap();

        search = &search[cmp::min(best_match.len(), search.len())..].trim_start();
        result.push((best_match.to_string(), *closure_map.get(best_type).unwrap()));
    }

    result
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_custom_patterns() {
        let custom_finnish = vec![
            ("viime [wday]", &Pattern::LastWday),
            ("edellinen [wday]", &Pattern::PrevWday),
            ("ensi [wday]", &Pattern::NextWday),
            ("seuraava [wday]", &Pattern::NextWday),
        ];

        let result_value = convert_custom("viime [wday]", vec![1], "2024-01-19T15:22:28+02:00", &custom_finnish);
        assert_eq!(result_value, "2024-01-15 00:00:00 +02:00");

        let result_value = convert_custom("edellinen [wday]", vec![1], "2024-01-19T15:22:28+02:00", &custom_finnish);
        assert_eq!(result_value, "2024-01-15 00:00:00 +02:00");

        let result_value = convert_custom("ensi [wday]", vec![1], "2024-01-19T15:22:28+02:00", &custom_finnish);
        assert_eq!(result_value, "2024-01-22 00:00:00 +02:00");

        let result_value = convert_custom("seuraava [wday]", vec![1], "2024-01-19T15:22:28+02:00", &custom_finnish);
        assert_eq!(result_value, "2024-01-22 00:00:00 +02:00");
    }

    fn convert_custom(pattern: &str, values: Vec<i64>, current_time: &str, custom: &Vec<(&str, &Pattern)>) -> String {
        let current_time = DateTime::parse_from_rfc3339(current_time).unwrap();
        let mut custom_patterns: HashMap<String, String> = HashMap::new();

        for (key, value) in custom {
            for pattern_value in Pattern::values(value) {
                custom_patterns.insert(key.to_string(), pattern_value.to_string());
            }
        }

        let tokens = values
            .iter()
            .map(|v| Token::new_integer(v.to_owned(), 0))
            .collect::<Vec<Token>>();

        let result_time = convert(pattern, tokens, &current_time, false, custom_patterns);
        result_time.unwrap().to_string()
    }
}
